# src/__init__.py
from .rebrandly_otel import *
from .flask_support import setup_flask
from .fastapi_support import setup_fastapi
from .pymysql_instrumentation import instrument_pymysql

# Explicitly define what's available
__all__ = [
    'otel',
    'lambda_handler',
    'span',
    'aws_message_span',
    'traces',
    'tracer',
    'metrics',
    'logger',
    'force_flush',
    'aws_message_handler',
    'shutdown',
    'setup_flask',
    'setup_fastapi',
    'instrument_pymysql'
]


# flask_integration.py
"""Flask integration for Rebrandly OTEL SDK."""

import json
from rebrandly_otel import Status, StatusCode, SpanKind

from flask import request, jsonify
from werkzeug.exceptions import HTTPException


def setup_flask(otel, app):
    """
    Setup Flask application with OTEL instrumentation.

    Example:
        from flask import Flask
        from rebrandly_otel import otel
        from rebrandly_otel.flask_integration import setup_flask
        
        app = Flask(__name__)
        setup_flask(otel, app)
    """
    app.before_request(lambda: app_before_request(otel))
    app.after_request(lambda response: app_after_request(otel, response))
    app.register_error_handler(Exception, lambda e: flask_error_handler(otel, e))
    return app

def app_before_request(otel):
    """
    Setup tracing for incoming Flask request.
    To be used with Flask's before_request hook.
    """

    # Extract trace context from headers
    headers = dict(request.headers)
    token = otel.attach_context(headers)
    request.trace_token = token

    # Determine span name - use route if available, otherwise just method
    # Route will be available after request routing is done
    span_name = f"{request.method} {request.path}"

    # Build attributes dict, excluding None values
    attributes = {
        # Required HTTP attributes per semantic conventions
        "http.request.method": request.method,
        "http.request.headers": json.dumps(request.headers, default=str),
        "url.full": request.url,
        "url.scheme": request.scheme,
        "url.path": request.path,
        "http.route": request.path,  # Flask doesn't expose route pattern easily
        "network.protocol.version": request.environ.get('SERVER_PROTOCOL', 'HTTP/1.1').split('/')[-1],
        "server.address": request.host.split(':')[0],
        "server.port": request.host.split(':')[1] if ':' in request.host else (443 if request.scheme == 'https' else 80),
    }

    # Add optional attributes only if they have values
    if request.query_string:
        attributes["url.query"] = request.query_string.decode('utf-8')

    if request.user_agent and request.user_agent.string:
        attributes["user_agent.original"] = request.user_agent.string

    if request.remote_addr:
        attributes["client.address"] = request.remote_addr

    # Start span for request using start_as_current_span to make it the active span
    span = otel.tracer.tracer.start_as_current_span(
        span_name,
        attributes=attributes,
        kind=SpanKind.SERVER
    )
    # Store both the span context manager and the span itself
    request.span_context = span
    request.span = span.__enter__()  # This activates the span and returns the span object

    # Log request start
    otel.logger.logger.info(f"Request started: {request.method} {request.path}",
                            extra={"http.method": request.method, "http.path": request.path})

def app_after_request(otel, response):
    """
    Cleanup tracing after Flask request completes.
    To be used with Flask's after_request hook.
    """

    # Check if we have a span and it's still recording
    if hasattr(request, 'span') and request.span.is_recording():
        # Update both new and old attribute names for compatibility
        request.span.set_attribute("http.response.status_code", response.status_code)

        # Set span status based on HTTP status code
        if response.status_code >= 400:
            request.span.set_status(Status(StatusCode.ERROR, f"HTTP {response.status_code}"))
        else:
            request.span.set_status(Status(StatusCode.OK))

        # Properly close the span context manager
        if hasattr(request, 'span_context'):
            request.span_context.__exit__(None, None, None)
        else:
            # Fallback if we don't have the context manager
            request.span.end()

    # Detach context
    if hasattr(request, 'trace_token'):
        otel.detach_context(request.trace_token)

    # Log request completion
    otel.logger.logger.info(f"Request completed: {response.status_code}",
                            extra={"http.status_code": response.status_code})

    otel.force_flush(timeout_millis=100)
    return response

def flask_error_handler(otel, exception):
    """
    Handle Flask exceptions and record them in the current span.
    To be used with Flask's errorhandler decorator.
    """

    # Determine the status code
    if isinstance(exception, HTTPException):
        status_code = exception.code
    elif hasattr(exception, 'status_code'):
        status_code = exception.status_code
    elif hasattr(exception, 'code'):
        status_code = exception.code if isinstance(exception.code, int) else 500
    else:
        status_code = 500

    # Record exception in span if available and still recording
    if hasattr(request, 'span') and request.span.is_recording():
        request.span.set_attribute("http.response.status_code", status_code)
        request.span.set_attribute("error.type", type(exception).__name__)

        request.span.record_exception(exception)
        request.span.set_status(Status(StatusCode.ERROR, str(exception)))
        request.span.add_event("exception", {
            "exception.type": type(exception).__name__,
            "exception.message": str(exception)
        })

        # Only close the span if it's still recording (not already ended)
        if hasattr(request, 'span_context'):
            request.span_context.__exit__(type(exception), exception, None)
        else:
            request.span.end()

    # Log the error with status code
    otel.logger.logger.error(f"Unhandled exception: {exception} (status: {status_code})",
                             exc_info=True,
                             extra={
                                 "exception.type": type(exception).__name__,
                                 "http.status_code": status_code
                             })

    # Return error response with the determined status code
    return jsonify({
        "error": str(exception),
        "type": type(exception).__name__
    }), status_code


# logs.py
"""Logging implementation for Rebrandly OTEL SDK."""
import logging
import sys
from typing import Optional
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import (
    BatchLogRecordProcessor,
    ConsoleLogExporter,
    SimpleLogRecordProcessor
)
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
from opentelemetry._logs import set_logger_provider

from .otel_utils import *


class RebrandlyLogger:
    """Wrapper for OpenTelemetry logging with Rebrandly-specific features."""

    def __init__(self):
        self._logger: Optional[logging.Logger] = None
        self._provider: Optional[LoggerProvider] = None
        self._setup_logging()

    def _setup_logging(self):
        """Initialize logging with configured exporters."""

        # Create provider with resource
        self._provider = LoggerProvider(resource=create_resource())

        # Add console exporter for local debugging
        if is_otel_debug():
            console_exporter = ConsoleLogExporter()
            self._provider.add_log_record_processor(SimpleLogRecordProcessor(console_exporter))

        # Add OTLP exporter if configured
        otel_endpoint = get_otlp_endpoint()
        if otel_endpoint:
            otlp_exporter = OTLPLogExporter(
                timeout=5,
                endpoint=otel_endpoint
            )
            batch_processor = BatchLogRecordProcessor(otlp_exporter, export_timeout_millis=get_millis_batch_time())
            self._provider.add_log_record_processor(batch_processor)

        set_logger_provider(self._provider)

        # Configure standard logging
        self._configure_standard_logging()

    def _configure_standard_logging(self):
        """Configure standard Python logging with OTEL handler."""
        # Get root logger
        root_logger = logging.getLogger()

        # Only configure basic logging if no handlers exist (not in Lambda)
        if not root_logger.handlers:
            logging.basicConfig(
                level=logging.INFO,
                format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                stream=sys.stdout
            )

        # Add OTEL handler without removing existing handlers
        otel_handler = LoggingHandler(logger_provider=self._provider)
        otel_handler.setLevel(logging.INFO)
        root_logger.addHandler(otel_handler)

        # Create service-specific logger
        self._logger = logging.getLogger(get_service_name())


    @property
    def logger(self) -> logging.Logger:
        """Get the standard Python logger."""
        if not self._logger:
            self._logger = logging.getLogger(get_service_name())
        return self._logger

    def force_flush(self, timeout_millis: int = 5000) -> bool:
        """
        Force flush all pending logs.

        Args:
            timeout_millis: Maximum time to wait for flush in milliseconds

        Returns:
            True if flush succeeded, False otherwise
        """
        if not self._provider:
            return True

        try:
            # Force flush the logger provider
            success = self._provider.force_flush(timeout_millis)

            # Also flush Python's logging handlers
            if self._logger:
                for handler in self._logger.handlers:
                    if hasattr(handler, 'flush'):
                        handler.flush()

            return success
        except Exception as e:
            print(f"[Logger] Error during force flush: {e}")
            return False

    def shutdown(self):
        """Shutdown the logger provider."""
        if self._provider:
            try:
                self._provider.shutdown()
                print("[Logger] Shutdown completed")
            except Exception as e:
                print(f"[Logger] Error during shutdown: {e}")



# traces.py
"""Tracing implementation for Rebrandly OTEL SDK."""
from typing import Optional, Dict, Any, ContextManager
from contextlib import contextmanager
from opentelemetry import trace, propagate, context
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace import TracerProvider, Span
from opentelemetry.sdk.trace.export import (
    ConsoleSpanExporter,
    BatchSpanProcessor,
    SimpleSpanProcessor
)

from .otel_utils import *

class RebrandlyTracer:
    """Wrapper for OpenTelemetry tracing with Rebrandly-specific features."""

    def __init__(self):
        self._tracer: Optional[trace.Tracer] = None
        self._provider: Optional[TracerProvider] = None
        self._setup_tracing()

    def _setup_tracing(self):

        # Create provider with resource
        self._provider = TracerProvider(resource=create_resource())

        # Add console exporter for local debugging
        if is_otel_debug():
            console_exporter = ConsoleSpanExporter()
            self._provider.add_span_processor(SimpleSpanProcessor(console_exporter))

        # Add OTLP exporter if configured
        otel_endpoint = get_otlp_endpoint()
        if otel_endpoint is not None:
            otlp_exporter = OTLPSpanExporter(
                endpoint=otel_endpoint,
                timeout=5
            )

            # Use batch processor for production
            batch_processor = BatchSpanProcessor(otlp_exporter, export_timeout_millis=get_millis_batch_time())
            self._provider.add_span_processor(batch_processor)

        # Set as global provider
        trace.set_tracer_provider(self._provider)

        # Get tracer
        self._tracer = trace.get_tracer(get_service_name(), get_service_version())

    def force_flush(self, timeout_millis: int = 5000) -> bool:
        """
        Force flush all pending spans.

        Args:
            timeout_millis: Maximum time to wait for flush in milliseconds

        Returns:
            True if flush succeeded, False otherwise
        """
        if not self._provider:
            return True

        try:
            # ForceFlush on the TracerProvider will flush all processors
            success = self._provider.force_flush(timeout_millis)

            if not success:
                print(f"[Tracer] Force flush timed out after {timeout_millis}ms")

            return success
        except Exception as e:
            print(f"[Tracer] Error during force flush: {e}")
            return False

    def shutdown(self):
        """Shutdown the tracer provider and all processors."""
        if self._provider:
            try:
                self._provider.shutdown()
                print("[Tracer] Shutdown completed")
            except Exception as e:
                print(f"[Tracer] Error during shutdown: {e}")

    @property
    def tracer(self) -> trace.Tracer:
        """Get the underlying OpenTelemetry tracer."""
        if not self._tracer:
            # Return no-op tracer if tracing is disabled
            return trace.get_tracer(__name__)
        return self._tracer

    @contextmanager
    def start_span(self,
                   name: str,
                   attributes: Optional[Dict[str, Any]] = None,
                   kind: trace.SpanKind = trace.SpanKind.INTERNAL) -> ContextManager[Span]:
        """Start a new span as the current span."""
        # Ensure we use the tracer to create a child span of the current span
        with self.tracer.start_as_current_span(
                name,
                attributes=attributes,
                kind=kind
        ) as span:
            yield span

    def start_span_with_context(self,
                                name: str,
                                attributes: Dict[str, str],
                                context_attributes: Optional[Dict[str, Any]] = None):
        """Start a span with extracted context (e.g., from message headers)."""
        # Extract context from carrier

        carrier, extracted_context = self.__get_aws_message_context_attributes(context_attributes)
        ctx = propagate.extract(extracted_context)

        if context_attributes is not None:
            # Start span with extracted context
            with self.tracer.start_as_current_span(
                    name,
                    context=ctx,
                    attributes=attributes
            ) as span:
                yield span
        else:
            # Start span with current context
            with self.tracer.start_as_current_span(
                    name,
                    context=context.get_current(),
                    attributes=attributes
            ) as span:
                yield span

    def get_current_span(self) -> Span:
        """Get the currently active span."""
        return trace.get_current_span()


    def record_span_exception(self, exception: Exception=None, span: Optional[Span] = None, msg: Optional[str] = None):
        """Record an exception on a span."""
        target_span = span or self.get_current_span()
        if target_span and hasattr(target_span, 'record_exception'):
            if exception is not None:
                target_span.record_exception(exception)
                target_span.set_status(trace.Status(trace.StatusCode.ERROR, str(exception)))


    def record_span_success(self, span: Optional[Span] = None, msg: Optional[str] = None):
        """Record success on a span."""
        target_span = span or self.get_current_span()
        if target_span and hasattr(target_span, 'set_status'):
            target_span.set_status(trace.Status(trace.StatusCode.OK))


    def add_event(self, name: str, attributes: Optional[Dict[str, Any]] = None, span: Optional[Span] = None):
        """Add an event to a span."""
        target_span = span or self.get_current_span()
        if target_span and hasattr(target_span, 'add_event'):
            target_span.add_event(name, attributes=attributes or {})

    # AWS-specific helpers
    def __get_aws_message_context_attributes(self, msg: dict):
        """
        Get trace context as AWS message attributes format.
        Used for SQS/SNS message propagation.
        """
        carrier = {}
        # Convert to AWS message attributes format
        message_attributes = {}
        if msg is not None and 'MessageAttributes' in msg:
            for key, value in msg['MessageAttributes'].items():
                carrier[key] = {
                    'StringValue': value,
                    'DataType': 'String'
                }
        context_extracted = propagate.extract(carrier)
        return carrier, context_extracted

    def get_attributes_for_aws_from_context(self):
        # Create carrier for message attributes
        carrier = {}

        # Inject trace context into carrier
        propagate.inject(carrier)

        # Convert carrier to SQS message attributes format
        message_attributes = {}
        for key, value in carrier.items():
            message_attributes[key] = {
                'StringValue': value,
                'DataType': 'String'
            }
        return message_attributes



# otel_utils.py

import os
import sys
import grpc
import json

from opentelemetry.sdk.resources import Resource, SERVICE_NAMESPACE
from opentelemetry.semconv.attributes import service_attributes
from opentelemetry.semconv._incubating.attributes import process_attributes, deployment_attributes

def create_resource(name: str = None, version: str = None) -> Resource:

    if name is None:
        name = get_service_name()
    if version is None:
        version = get_service_version()

    resource = Resource.create(
        {
            service_attributes.SERVICE_NAME: name,
            service_attributes.SERVICE_VERSION: version,
            process_attributes.PROCESS_RUNTIME_VERSION: f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            SERVICE_NAMESPACE: os.environ.get('ENV', os.environ.get('ENVIRONMENT', os.environ.get('NODE_ENV', 'local')))
        }
    )
    return resource

def get_package_version():
    try:
        from importlib.metadata import version, PackageNotFoundError  # Python 3.8+
        return version('rebrandly_otel')
    except ImportError:
        try:
            from importlib_metadata import version, PackageNotFoundError
            return version('rebrandly_otel')
        except Exception as e:
            print(f"[OTEL Utils] Warning: Could not get package version: {e}")
            return '0.1.0'


def get_service_name(service_name: str = None) -> str:
    if service_name is None:
        serv = os.environ.get('OTEL_SERVICE_NAME', 'default-service-python')
        if serv.strip() == "":
            return 'default-service-python'
        return serv
    return service_name


def get_service_version(service_version: str = None) -> str:
    if service_version is None:
        return os.environ.get('OTEL_SERVICE_VERSION', get_package_version())
    return service_version


def get_otlp_endpoint(otlp_endpoint: str = None) -> str | None:
    endpoint = otlp_endpoint or os.environ.get('OTEL_EXPORTER_OTLP_ENDPOINT', None)

    if endpoint.strip() == "":
        return None

    if endpoint is not None:
        try:
            from urllib.parse import urlparse

            # Parse the endpoint
            parsed = urlparse(endpoint if '://' in endpoint else f'http://{endpoint}')
            host = parsed.hostname
            port = parsed.port

            # Test gRPC connection
            channel = grpc.insecure_channel(f'{host}:{port}')
            try:
                # Wait for the channel to be ready
                grpc.channel_ready_future(channel).result(timeout=3)
                return endpoint
            finally:
                channel.close()

        except Exception as e:
            print(f"[OTEL] Failed to connect to OTLP endpoint {endpoint}: {e}")
            return None
    return endpoint

def is_otel_debug() -> bool:
    return os.environ.get('OTEL_DEBUG', 'false').lower() == 'true'


def get_millis_batch_time():
    try:
        return int(os.environ.get('BATCH_EXPORT_TIME_MILLIS', 100))
    except Exception as e:
        print(f"[OTEL Utils] Warning: Invalid BATCH_EXPORT_TIME_MILLIS value, using default 5000ms: {e}")
        return 5000

def extract_event_from(message) -> str | None:
    body = None
    if 'body' in message:
        body = message['body']
    if 'Body' in message:
        body = message['Body']
    if 'Message' in message:
        body = message['Message']
    if 'Sns' in message and 'Message' in message['Sns']:
        body = message['Sns']['Message']
    if body is not None:
        try:
            jbody = json.loads(body)
            if 'event' in jbody:
                return jbody['event']
        except:
            pass
    return None


# metrics.py
"""Metrics implementation for Rebrandly OTEL SDK."""
from typing import Optional, Dict, List
from dataclasses import dataclass
from enum import Enum
from opentelemetry import metrics
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.metrics import Meter, Histogram, Instrument, Counter
from opentelemetry.metrics._internal import Gauge
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import (PeriodicExportingMetricReader, ConsoleMetricExporter)
from opentelemetry.sdk.metrics._internal.aggregation import (ExplicitBucketHistogramAggregation)
from opentelemetry.sdk.metrics.view import View

from .otel_utils import *

class MetricType(Enum):
    """Supported metric types."""
    COUNTER = "counter"
    GAUGE = "gauge"
    HISTOGRAM = "histogram"
    UP_DOWN_COUNTER = "up_down_counter"

@dataclass
class MetricDefinition:
    """Definition of a metric."""
    name: str
    description: str
    unit: str = "1"
    type: MetricType = MetricType.COUNTER

class RebrandlyMeter:
    """Wrapper for OpenTelemetry metrics with Rebrandly-specific features."""

    # Standardized metric definitions aligned with Node.js
    DEFAULT_METRICS = {
        ## PROCESS
        'cpu_usage_percentage': MetricDefinition(
            name='process.cpu.utilization',
            description='Difference in process.cpu.time since the last measurement, divided by the elapsed time and number of CPUs available to the process.',
            unit='1',
            type=MetricType.GAUGE
        ),
        'memory_usage_bytes': MetricDefinition(
            name='process.memory.used',
            description='The amount of physical memory in use.',
            unit='By',
            type=MetricType.GAUGE
        )
    }

    class GlobalMetrics:
        def __init__(self, rebrandly_meter):
            self.__rebrandly_meter = rebrandly_meter
            self.cpu_usage_percentage: Gauge = self.__rebrandly_meter.get_metric('cpu_usage_percentage')
            self.memory_usage_bytes: Gauge = self.__rebrandly_meter.get_metric('memory_usage_bytes')


    def __init__(self):
        self._meter: Optional[Meter] = None
        self._provider: Optional[MeterProvider] = None
        self._metrics: Dict[str, Instrument] = {}
        self.__setup_metrics()
        self.__register_default_metrics()
        self.GlobalMetrics = RebrandlyMeter.GlobalMetrics(self)

    def __setup_metrics(self):
        """Initialize metrics with configured exporters."""

        readers = []

        # Add console exporter for local debugging
        if is_otel_debug():
            console_reader = PeriodicExportingMetricReader(
                ConsoleMetricExporter(),
                export_interval_millis=1000  # 10 seconds for debugging
            )
            readers.append(console_reader)

        # Add OTLP exporter if configured
        otel_endpoint = get_otlp_endpoint()
        if otel_endpoint is not None:
            otlp_exporter = OTLPMetricExporter(
                endpoint=otel_endpoint,
                timeout=5
            )
            otlp_reader = PeriodicExportingMetricReader(otlp_exporter, export_interval_millis=get_millis_batch_time())
            readers.append(otlp_reader)

        # Create views
        views = self.__create_views()

        # Create provider
        self._provider = MeterProvider(
            resource=create_resource(),
            metric_readers=readers,
            views=views
        )

        # Set as global provider
        metrics.set_meter_provider(self._provider)

        # Get meter
        self._meter = metrics.get_meter(get_service_name(), get_service_version())

    def __create_views(self) -> List[View]:
        """Create metric views for customization."""
        views = []

        # Histogram view with custom buckets
        histogram_view = View(
            instrument_type=Histogram,
            instrument_name="*",
            aggregation=ExplicitBucketHistogramAggregation((0.001, 0.004, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5)) # todo <-- define buckets
        )
        views.append(histogram_view)

        return views

    def __register_default_metrics(self):
        """Register default metrics."""
        for name, definition in self.DEFAULT_METRICS.items():
            self.register_metric(definition)

    @property
    def meter(self) -> Meter:
        """Get the underlying OpenTelemetry meter."""
        if not self._meter:
            # Return no-op meter if metrics are disabled
            return metrics.get_meter(__name__)
        return self._meter

    def force_flush(self, timeout_millis: int = 5000) -> bool:
        """
        Force flush all pending metrics.

        Args:
            timeout_millis: Maximum time to wait for flush in milliseconds

        Returns:
            True if flush succeeded, False otherwise
        """
        if not hasattr(self, '_provider') or not self._provider:
            return True

        try:
            # Get the internal provider (MeterProvider doesn't have direct flush)
            # We need to flush through the metric readers
            success = self._provider.force_flush(timeout_millis)
            return success
        except Exception as e:
            print(f"[Meter] Error during force flush: {e}")
            # For metrics, we might not have a flush method, so we return True
            return True

    def shutdown(self):
        """Shutdown the meter provider."""
        if hasattr(self, '_provider') and self._provider:
            try:
                self._provider.shutdown()
                print("[Meter] Shutdown completed")
            except Exception as e:
                print(f"[Meter] Error during shutdown: {e}")

    def register_metric(self, definition: MetricDefinition) -> Instrument:
        """Register a new metric."""
        if definition.name in self._metrics:
            return self._metrics[definition.name]

        metric = self.__create_metric(definition)
        self._metrics[definition.name] = metric
        return metric

    def __create_metric(self, definition: MetricDefinition) -> Instrument:
        """Create a metric instrument based on definition."""
        if definition.type == MetricType.COUNTER:
            return self.meter.create_counter(
                name=definition.name,
                unit=definition.unit,
                description=definition.description
            )
        elif definition.type == MetricType.HISTOGRAM:
            return self.meter.create_histogram(
                name=definition.name,
                unit=definition.unit,
                description=definition.description
            )
        elif definition.type == MetricType.UP_DOWN_COUNTER:
            return self.meter.create_up_down_counter(
                name=definition.name,
                unit=definition.unit,
                description=definition.description
            )
        elif definition.type == MetricType.GAUGE:
            # For gauges, we'll create them when needed with callbacks
            return self.meter.create_gauge(
                name=definition.name,
                unit=definition.unit,
                description=definition.description
            )
        else:
            raise ValueError(f"Unknown metric type: {definition.type}")

    def get_metric(self, name: str) -> Optional[Instrument]:
        """Get a registered metric by name."""
        return self._metrics.get(name)



import unittest
from unittest.mock import patch, MagicMock, ANY, call
import importlib
import os

# Import the module under test
import src.rebrandly_otel
from src.rebrandly_otel import RebrandlyOTEL, otel, lambda_handler, aws_message_handler, span, force_flush, shutdown
from opentelemetry.trace import SpanKind, Status, StatusCode


class TestRebrandlyOTELUsage(unittest.TestCase):

    def setUp(self):
        # Reset the singleton and reload the module before each test
        RebrandlyOTEL._instance = None
        RebrandlyOTEL._initialized = False
        importlib.reload(src.rebrandly_otel)

        # Patch external dependencies at the module level
        self.mock_propagate = patch('opentelemetry.propagate').start()
        self.mock_context = patch('opentelemetry.context').start()
        self.mock_psutil = patch('src.rebrandly_otel.psutil').start()

        # Patch the attributes of the otel instance directly
        self.mock_tracer = MagicMock()
        self.mock_meter = MagicMock()
        self.mock_logger = MagicMock()

        otel._tracer = self.mock_tracer
        otel._meter = self.mock_meter
        otel._logger = self.mock_logger

        # Configure mock behaviors for tracer
        mock_span_instance = MagicMock()
        mock_span_instance.set_status = MagicMock()
        mock_span_instance.record_exception = MagicMock()
        mock_span_instance.set_attribute = MagicMock()

        mock_start_span_context_manager = MagicMock()
        mock_start_span_context_manager.__enter__.return_value = mock_span_instance
        mock_start_span_context_manager.__exit__.return_value = None

        self.mock_tracer.start_span.return_value = mock_start_span_context_manager

        # Configure mock behaviors for meter
        self.mock_meter.GlobalMetrics.memory_usage_bytes.set = MagicMock()
        self.mock_meter.GlobalMetrics.cpu_usage_percentage.set = MagicMock()

        # Configure mock behaviors for psutil
        self.mock_psutil.cpu_percent.return_value = 10.0
        mock_virtual_memory = MagicMock()
        mock_virtual_memory.used = 1000000
        mock_virtual_memory.percent = 50.0
        self.mock_psutil.virtual_memory.return_value = mock_virtual_memory

        self.addCleanup(patch.stopall)

    def test_initialization(self):
        # otel instance is already created and initialized in setUp
        self.assertIsInstance(otel, RebrandlyOTEL)
        # The tracer, meter, logger properties are accessed during otel initialization
        # which happens when the module is reloaded in setUp.
        # So, we check that the mocks were assigned.
        self.assertIs(otel.tracer, self.mock_tracer)
        self.assertIs(otel.meter, self.mock_meter)
        self.assertIs(otel.logger, self.mock_logger)

    def test_span_context_manager(self):
        with span("test_span", attributes={"key": "value"}) as s:
            self.assertIsNotNone(s)
            s.set_attribute("another_key", "another_value")

        self.mock_tracer.start_span.assert_called_once_with(
            "test_span", attributes={"key": "value"}, kind=SpanKind.INTERNAL
        )
        span_mock = self.mock_tracer.start_span.return_value.__enter__.return_value
        self.assertEqual(span_mock.set_status.call_args[0][0].status_code, StatusCode.OK)

    def test_span_context_manager_with_exception(self):
        with self.assertRaises(ValueError):
            with span("error_span"):
                raise ValueError("Something went wrong")

        self.mock_tracer.start_span.assert_called_once_with(
            "error_span", attributes=None, kind=SpanKind.INTERNAL
        )
        span_mock = self.mock_tracer.start_span.return_value.__enter__.return_value
        span_mock.record_exception.assert_called_once()
        self.assertEqual(span_mock.set_status.call_args[0][0].status_code, StatusCode.ERROR)
        self.assertEqual(span_mock.set_status.call_args[0][0].description, "Something went wrong")

    def test_lambda_handler_basic(self):
        @lambda_handler("my_lambda")
        def handler(event, context):
            return {"statusCode": 200, "body": "OK"}

        mock_context = MagicMock()
        mock_context.aws_request_id = "req123"
        mock_context.function_arn = "arn:aws:lambda:us-east-1:123456789012:function:my_lambda"
        mock_context.function_name = "my_lambda"
        mock_context.function_version = "$LATEST"

        result = handler({}, mock_context)

        self.assertEqual(result, {"statusCode": 200, "body": "OK"})
        self.mock_tracer.start_span.assert_called_once_with(
            name="my_lambda",
            attributes={
                "faas.trigger": "direct",
                "faas.execution": "req123",
                "faas.id": "arn:aws:lambda:us-east-1:123456789012:function:my_lambda",
                "faas.name": "my_lambda",
                "faas.version": "$LATEST"
            },
            kind=SpanKind.SERVER
        )
        span_mock = self.mock_tracer.start_span.return_value.__enter__.return_value
        self.assertEqual(span_mock.set_status.call_args[0][0].status_code, StatusCode.OK)
        self.mock_logger.logger.info.assert_called_with(ANY)
        self.mock_psutil.cpu_percent.assert_called_once()
        self.mock_psutil.virtual_memory.assert_called_once()
        self.mock_meter.GlobalMetrics.memory_usage_bytes.set.assert_called_once()
        self.mock_meter.GlobalMetrics.cpu_usage_percentage.set.assert_called_once()
        self.mock_tracer.force_flush.assert_called_once()
        self.mock_meter.force_flush.assert_called_once()
        self.mock_logger.force_flush.assert_called_once()

    def test_lambda_handler_with_sqs_event(self):
        @lambda_handler("sqs_processor")
        def handler(event, context):
            return {"statusCode": 200}

        sqs_event = {
            "Records": [
                {
                    "messageId": "msg1",
                    "eventSource": "aws:sqs",
                    "messageAttributes": {
                        "traceparent": {"stringValue": "00-0123456789abcdef0123456789abcdef-0123456789abcdef-01", "dataType": "String"}
                    }
                }
            ]
        }
        mock_context = MagicMock()
        mock_context.aws_request_id = "req456"

        handler(sqs_event, mock_context)

        self.mock_propagate.extract.assert_called_once_with({'traceparent': '00-0123456789abcdef0123456789abcdef-0123456789abcdef-01'})
        self.mock_context.attach.assert_called_once()
        self.mock_context.detach.assert_called_once()
        self.mock_tracer.start_span.assert_called_once_with(
            name="sqs_processor",
            attributes=ANY,
            kind=SpanKind.SERVER
        )
        span_mock = self.mock_tracer.start_span.return_value.__enter__.return_value
        self.assertEqual(span_mock.set_status.call_args[0][0].status_code, StatusCode.OK)
        self.mock_logger.logger.info.assert_called_with(ANY)
        self.mock_psutil.cpu_percent.assert_called_once()
        self.mock_psutil.virtual_memory.assert_called_once()
        self.mock_meter.GlobalMetrics.memory_usage_bytes.set.assert_called_once()
        self.mock_meter.GlobalMetrics.cpu_usage_percentage.set.assert_called_once()
        self.mock_tracer.force_flush.assert_called_once()
        self.mock_meter.force_flush.assert_called_once()
        self.mock_logger.force_flush.assert_called_once()

    def test_lambda_handler_with_sns_event(self):
        @lambda_handler("sns_processor")
        def handler(event, context):
            return {"statusCode": 200}

        sns_event = {
            "Records": [
                {
                    "EventSource": "aws:sns",
                    "Sns": {
                        "MessageId": "sns_msg1",
                        "TopicArn": "arn:aws:sns:us-east-1:123456789012:my-topic",
                        "Subject": "test_subject",
                        "MessageAttributes": {
                            "traceparent": {"Type": "String", "Value": "00-fedcba9876543210fedcba9876543210-fedcba9876543210-01"}
                        }
                    }
                }
            ]
        }
        mock_context = MagicMock()
        mock_context.aws_request_id = "req789"

        handler(sns_event, mock_context)

        self.mock_propagate.extract.assert_called_once_with({'traceparent': '00-fedcba9876543210fedcba9876543210-fedcba9876543210-01'})
        self.mock_context.attach.assert_called_once()
        self.mock_context.detach.assert_called_once()
        self.mock_tracer.start_span.assert_called_once_with(
            name="sns_processor",
            attributes=ANY,
            kind=SpanKind.SERVER
        )
        span_mock = self.mock_tracer.start_span.return_value.__enter__.return_value
        self.assertEqual(span_mock.set_status.call_args[0][0].status_code, StatusCode.OK)
        self.mock_logger.logger.info.assert_called_with(ANY)
        self.mock_psutil.cpu_percent.assert_called_once()
        self.mock_psutil.virtual_memory.assert_called_once()
        self.mock_meter.GlobalMetrics.memory_usage_bytes.set.assert_called_once()
        self.mock_meter.GlobalMetrics.cpu_usage_percentage.set.assert_called_once()
        self.mock_tracer.force_flush.assert_called_once()
        self.mock_meter.force_flush.assert_called_once()
        self.mock_logger.force_flush.assert_called_once()

    def test_aws_message_handler_basic(self):
        @aws_message_handler("msg_processor")
        def handler(record):
            return {"processed": True}

        record = {"messageId": "rec123"}
        result = handler(record)

        self.assertEqual(result, {"processed": True})
        self.mock_tracer.start_span.assert_called_once_with(
            name="msg_processor",
            attributes={'messaging.operation': 'process'},
            kind=SpanKind.CONSUMER
        )
        span_mock = self.mock_tracer.start_span.return_value.__enter__.return_value
        self.assertEqual(span_mock.set_status.call_args[0][0].status_code, StatusCode.OK)
        self.mock_psutil.cpu_percent.assert_called_once()
        self.mock_psutil.virtual_memory.assert_called_once()
        self.mock_meter.GlobalMetrics.memory_usage_bytes.set.assert_called_once()
        self.mock_meter.GlobalMetrics.cpu_usage_percentage.set.assert_called_once()
        self.mock_tracer.force_flush.assert_called_once()
        self.mock_meter.force_flush.assert_called_once()
        self.mock_logger.force_flush.assert_called_once()

    def test_aws_message_span_context_manager(self):
        sqs_message = {
            "messageId": "sqs_msg_id",
            "eventSource": "aws:sqs",
            "awsRegion": "us-east-1",
            "messageAttributes": {
                "traceparent": {"stringValue": "00-11111111111111111111111111111111-2222222222222222-01", "dataType": "String"}
            }
        }
        with otel.aws_message_span("process_sqs", message=sqs_message) as s:
            self.assertIsNotNone(s)
            s.set_attribute("custom_attr", "custom_value")

        self.mock_tracer.start_span.assert_called_once_with(
            name="process_sqs",
            attributes={
                'messaging.operation': 'process',
                'messaging.message_id': 'sqs_msg_id',
                'messaging.system': 'aws:sqs',
                'cloud.region': 'us-east-1'
            },
            kind=SpanKind.CONSUMER
        )
        span_mock = self.mock_tracer.start_span.return_value.__enter__.return_value
        span_mock.set_attribute.assert_called_once_with("custom_attr", "custom_value")
        self.assertEqual(span_mock.set_status.call_args[0][0].status_code, StatusCode.OK)

    def test_force_flush(self):
        otel.force_flush()
        self.mock_tracer.force_flush.assert_called_once()
        self.mock_meter.force_flush.assert_called_once()
        self.mock_logger.force_flush.assert_called_once()

    def test_shutdown(self):
        otel.shutdown()
        self.mock_tracer.shutdown.assert_called_once()
        self.mock_meter.shutdown.assert_called_once()
        self.mock_logger.shutdown.assert_called_once()

    def test_detect_lambda_trigger(self):
        # Test direct invocation
        self.assertEqual(otel._detect_lambda_trigger(None), 'direct')
        self.assertEqual(otel._detect_lambda_trigger({}), 'direct')

        # Test SQS
        sqs_event = {'Records': [{'eventSource': 'aws:sqs'}]}
        self.assertEqual(otel._detect_lambda_trigger(sqs_event), 'sqs')

        # Test SNS
        sns_event = {'Records': [{'eventSource': 'aws:sns'}]}
        self.assertEqual(otel._detect_lambda_trigger(sns_event), 'sns')

        # Test API Gateway
        api_gw_event = {'httpMethod': 'GET'}
        self.assertEqual(otel._detect_lambda_trigger(api_gw_event), 'api_gateway')

        api_gw_v2_event = {'requestContext': {'http': {}}}
        self.assertEqual(otel._detect_lambda_trigger(api_gw_v2_event), 'api_gateway_v2')

        # Test EventBridge
        eventbridge_event = {'source': 'aws.events'}
        self.assertEqual(otel._detect_lambda_trigger(eventbridge_event), 'eventbridge')

        # Test unknown
        unknown_event = {'some_key': 'some_value'}
        self.assertEqual(otel._detect_lambda_trigger(unknown_event), 'unknown')

    def test_context_propagation_methods(self):
        # Mock the global propagate and context objects
        mock_propagate = MagicMock()
        mock_context = MagicMock()

        with patch('src.rebrandly_otel.propagate', mock_propagate), patch('src.rebrandly_otel.context', mock_context):

            # Test inject_context
            carrier = {}
            otel.inject_context(carrier)
            mock_propagate.inject.assert_called_once_with(carrier)

            # Test extract_context
            extracted_ctx = otel.extract_context(carrier)
            mock_propagate.extract.assert_called_once_with(carrier)
            self.assertEqual(extracted_ctx, mock_propagate.extract.return_value)

            # Test attach_context
            mock_propagate.extract.reset_mock()
            mock_context.attach.reset_mock()
            token = otel.attach_context(carrier)
            mock_propagate.extract.assert_called_once_with(carrier)
            mock_context.attach.assert_called_once_with(mock_propagate.extract.return_value)
            self.assertEqual(token, mock_context.attach.return_value)

            # Test detach_context
            mock_context.detach.reset_mock()
            otel.detach_context(token)
            mock_context.detach.assert_called_once_with(token)


if __name__ == '__main__':
    unittest.main()



import json
import time
import psutil
import functools
from contextlib import contextmanager
from datetime import datetime
from opentelemetry.trace import Status, StatusCode, SpanKind
from typing import Optional, Dict, Any, Callable, TypeVar
from opentelemetry import baggage, propagate, context

from .traces import RebrandlyTracer
from .metrics import RebrandlyMeter
from .logs import RebrandlyLogger
from .otel_utils import extract_event_from


T = TypeVar('T')

class RebrandlyOTEL:
    """Main entry point for Rebrandly's OpenTelemetry instrumentation."""

    _instance: Optional['RebrandlyOTEL'] = None
    _initialized: bool = False

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        if not self._initialized:
            self._tracer: Optional[RebrandlyTracer] = None
            self._meter: Optional[RebrandlyMeter] = None
            self._logger: Optional[RebrandlyLogger] = None
            self.__class__._initialized = True

    def initialize(self, **kwargs) -> 'RebrandlyOTEL':
        # Force initialization of components
        _ = self.tracer
        _ = self.meter
        _ = self.logger

        return self

    @property
    def tracer(self) -> RebrandlyTracer:
        """Get the tracer instance."""
        if self._tracer is None:
            self._tracer = RebrandlyTracer()
        return self._tracer

    @property
    def meter(self) -> RebrandlyMeter:
        """Get the meter instance."""
        if self._meter is None:
            self._meter = RebrandlyMeter()
        return self._meter

    @property
    def logger(self) -> RebrandlyLogger:
        """Get the logger instance."""
        if self._logger is None:
            self._logger = RebrandlyLogger()
        return self._logger

    # Convenience methods for common operations

    @contextmanager
    def span(self,
             name: str,
             attributes: Optional[Dict[str, Any]] = None,
             kind: SpanKind = SpanKind.INTERNAL,
             message=None):
        """Create a span using context manager."""
        with self.tracer.start_span(name, attributes=attributes, kind=kind) as span:
            try:
                yield span
                span.set_status(Status(StatusCode.OK))
            except Exception as e:
                span.record_exception(e)
                span.set_status(Status(StatusCode.ERROR, str(e)))
                raise

    def trace_decorator(self,
                        name: Optional[str] = None,
                        attributes: Optional[Dict[str, Any]] = None,
                        kind: SpanKind = SpanKind.INTERNAL) -> Callable[[T], T]:
        """Decorator for tracing functions."""
        def decorator(func: T) -> T:
            span_name = name or f"{func.__module__}.{func.__name__}"

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                with self.span(span_name, attributes=attributes, kind=kind):
                    return func(*args, **kwargs)

            return wrapper
        return decorator

    def lambda_handler(self,
                       name: Optional[str] = None,
                       attributes: Optional[Dict[str, Any]] = None,
                       kind: SpanKind = SpanKind.SERVER,
                       auto_flush: bool = True,
                       skip_aws_link: bool = False):
        """
        Decorator specifically for Lambda handlers with automatic flushing.
        """
        def decorator(func):
            @functools.wraps(func)
            def wrapper(event=None, lambda_context=None):
                # Determine span name
                span_name = name or f"lambda.{func.__name__}"
                start_time = datetime.now()

                # Build span attributes
                span_attributes = attributes or {}
                span_attributes['faas.trigger'] = self._detect_lambda_trigger(event)

                # Add Lambda-specific attributes if context is available
                if lambda_context is not None:
                    span_attributes.update({
                        "faas.execution": getattr(lambda_context, 'aws_request_id', 'unknown'),
                        "faas.id": getattr(lambda_context, 'function_arn', 'unknown'),
                        "faas.name": getattr(lambda_context, 'function_name', 'unknown'),
                        "faas.version": getattr(lambda_context, 'function_version', 'unknown')
                    })

                # Handle context extraction from AWS events
                token = None

                if not skip_aws_link and event and isinstance(event, dict) and 'Records' in event:
                    first_record = event['Records'][0] if event['Records'] else None
                    if first_record:
                        carrier = {}

                        # Extract from SQS
                        if 'MessageAttributes' in first_record:
                            for key, value in first_record['MessageAttributes'].items():
                                if isinstance(value, dict) and 'StringValue' in value:
                                    carrier[key] = value['StringValue']
                        if ('messageAttributes' in first_record and 'traceparent' in first_record['messageAttributes']
                                and 'stringValue' in first_record['messageAttributes']['traceparent']):
                            carrier['traceparent'] = first_record['messageAttributes']['traceparent']['stringValue']

                        # Extract from SNS
                        elif 'Sns' in first_record and 'MessageAttributes' in first_record['Sns']:
                            for key, value in first_record['Sns']['MessageAttributes'].items():
                                if isinstance(value, dict):
                                    if 'Value' in value:
                                        carrier[key] = value['Value']
                                    elif 'StringValue' in value:
                                        carrier[key] = value['StringValue']

                        # Attach extracted context
                        if carrier:
                            from opentelemetry import propagate, context as otel_context
                            extracted_context = propagate.extract(carrier)
                            token = otel_context.attach(extracted_context)

                result = None
                try:

                    # Create and execute within span
                    with self.tracer.start_span(
                            name=span_name,
                            attributes=span_attributes,
                            kind=kind
                    ) as span:
                        # Add invocation start event with standardized attributes
                        start_event_attrs = {
                            'event.type': type(event).__name__ if event else 'None'
                        }

                        # Add records count if present
                        if event and isinstance(event, dict) and 'Records' in event:
                            start_event_attrs['event.records'] = f"{len(event['Records'])}"

                        span.add_event("lambda.invocation.start", start_event_attrs)

                        # Execute handler
                        result = func(event, lambda_context)

                        # Process result with standardized attributes
                        success = True
                        complete_event_attrs = {}

                        if isinstance(result, dict) and 'statusCode' in result:
                            span.set_attribute("http.status_code", result['statusCode'])
                            complete_event_attrs['status_code'] = result['statusCode']

                            # Set span status based on HTTP status code
                            if result['statusCode'] >= 400:
                                success = False
                                span.set_status(Status(StatusCode.ERROR, f"HTTP {result['statusCode']}"))
                            else:
                                span.set_status(Status(StatusCode.OK))
                        else:
                            span.set_status(Status(StatusCode.OK))

                        # Add completion event with success indicator
                        complete_event_attrs['success'] = success
                        span.add_event("lambda.invocation.complete", complete_event_attrs)

                    return result

                except Exception as e:

                    # Add failed completion event with error attribute (THIS IS THE KEY ADDITION)
                    span.add_event("lambda.invocation.complete", {
                        'success': False,
                        'error': type(e).__name__
                    })

                    # Record the exception in the span
                    span.record_exception(e)
                    span.set_status(Status(StatusCode.ERROR, str(e)))

                    # Log error
                    self.logger.logger.error(f"Lambda execution failed: {e}", exc_info=True)
                    raise

                finally:
                    # Always detach context if we attached it
                    if token is not None:
                        from opentelemetry import context as otel_context
                        otel_context.detach(token)

                    # Force flush if enabled
                    if auto_flush:
                        self.logger.logger.info(f"[Rebrandly OTEL] Lambda '{span_name}', flushing...")
                        flush_success = self.force_flush(timeout_millis=1000)
                        if not flush_success:
                            self.logger.logger.warning("[Rebrandly OTEL] Force flush may not have completed fully")

            return wrapper
        return decorator

    def aws_message_handler(self,
                            name: Optional[str] = None,
                            attributes: Optional[Dict[str, Any]] = None,
                            kind: SpanKind = SpanKind.CONSUMER,
                            auto_flush: bool = True):
        """
        Decorator for AWS message handlers (SQS/SNS record processing).
        Requires a record object parameter to the function.
        """
        def decorator(func):
            @functools.wraps(func)
            def wrapper(record=None, *args, **kwargs):
                # Determine span name
                span_name = name or f"message.{func.__name__}"
                start_func = datetime.now()

                # Build span attributes
                span_attributes = attributes or {}
                span_attributes['messaging.operation'] = 'process'

                result = None
                try:
                    # Create span and execute function
                    span_function = self.span
                    if record is not None and (('MessageAttributes' in record or 'messageAttributes' in record) or ('Sns' in record and 'MessageAttributes' in record['Sns'])):
                        span_function = self.aws_message_span
                        evt = extract_event_from(record)
                        if evt:
                            span_attributes['event.type'] = evt

                    with span_function(span_name, message=record, attributes=span_attributes, kind=kind) as span_context:
                        # Add processing start event with standardized name
                        span_context.add_event("message.processing.start", {})

                        # Execute the actual handler function
                        result = func(record, *args, **kwargs)

                        # Process result
                        success = True
                        complete_event_attrs = {}

                        if result and isinstance(result, dict):
                            if 'statusCode' in result:
                                span_context.set_attribute("http.status_code", result['statusCode'])

                                # Set span status based on status code
                                if result['statusCode'] >= 400:
                                    success = False
                                    span_context.set_status(
                                        Status(StatusCode.ERROR, f"Handler returned {result['statusCode']}")
                                    )
                                else:
                                    span_context.set_status(Status(StatusCode.OK))

                            # Add custom result attributes if present
                            if 'processed' in result:
                                complete_event_attrs['processed'] = result['processed']
                                span_context.set_attribute("message.processed", result['processed'])
                            if 'skipped' in result:
                                complete_event_attrs['skipped'] = result['skipped']
                                span_context.set_attribute("message.skipped", result['skipped'])
                        else:
                            span_context.set_status(Status(StatusCode.OK))

                        # Add completion event with standardized name
                        complete_event_attrs['success'] = success
                        span_context.add_event("message.processing.complete", complete_event_attrs)

                        return result

                except Exception as e:
                    # Record the exception in the span
                    if 'span_context' in locals():
                        span_context.record_exception(e)
                        span_context.set_status(Status(StatusCode.ERROR, str(e)))

                        # Add failed processing event
                        span_context.add_event("message.processing.complete", {
                            'success': False,
                            'error': type(e).__name__
                        })

                    # Re-raise the exception
                    raise

                finally:
                    if auto_flush:
                        self.force_flush(start_datetime=start_func)

            return wrapper
        return decorator

    def force_flush(self, start_datetime: datetime=None, timeout_millis: int = 1000) -> bool:
        """
        Force flush all telemetry data.
        This is CRITICAL for Lambda functions to ensure data is sent before function freezes.

        Args:
            start_datetime: Optional start time for system metrics capture
            timeout_millis: Maximum time to wait for flush in milliseconds

        Returns:
            True if all flushes succeeded, False otherwise
        """
        success = True

        if start_datetime is not None:
            end_func = datetime.now()
            cpu_percent = psutil.cpu_percent(interval=0.1)  # Shorter interval for Lambda
            memory = psutil.virtual_memory()

            # Record metrics using standardized names
            self.meter.GlobalMetrics.memory_usage_bytes.set(memory.used)
            self.meter.GlobalMetrics.cpu_usage_percentage.set(cpu_percent)

            print(f"Function Memory usage: {memory.percent}%, CPU usage: {cpu_percent}%")

        try:
            # Flush traces
            if self._tracer:
                if not self._tracer.force_flush(timeout_millis):
                    success = False

            # Flush metrics
            if self._meter:
                if not self._meter.force_flush(timeout_millis):
                    success = False

            # Flush logs
            if self._logger:
                if not self._logger.force_flush(timeout_millis):
                    success = False

            # Add a small delay to ensure network operations complete
            time.sleep(0.1)

        except Exception as e:
            print(f"[Rebrandly OTEL] Error during force flush: {e}")
            success = False

        return success

    def shutdown(self):
        """
        Shutdown all OTEL components gracefully.
        Call this at the end of your Lambda handler if you want to ensure clean shutdown.
        """
        try:
            if self._tracer:
                self._tracer.shutdown()
            if self._meter:
                self._meter.shutdown()
            if self._logger:
                self._logger.shutdown()
        except Exception as e:
            print(f"[Rebrandly OTEL] Error during shutdown: {e}")

    def _detect_lambda_trigger(self, event: Any) -> str:
        """Detect Lambda trigger type from event."""
        if not event or not isinstance(event, dict):
            return 'direct'

        if 'Records' in event:
            first_record = event['Records'][0] if event['Records'] else None
            if first_record:
                event_source = first_record.get('eventSource', '')
                if event_source == 'aws:sqs':
                    return 'sqs'
                elif event_source == 'aws:sns':
                    return 'sns'
                elif event_source == 'aws:s3':
                    return 's3'
                elif event_source == 'aws:kinesis':
                    return 'kinesis'
                elif event_source == 'aws:dynamodb':
                    return 'dynamodb'

        if 'httpMethod' in event:
            return 'api_gateway'
        if 'requestContext' in event and 'http' in event.get('requestContext', {}):
            return 'api_gateway_v2'
        if event.get('source') == 'aws.events':
            return 'eventbridge'
        if event.get('source') == 'aws.scheduler':
            return 'eventbridge_scheduler'
        if 'jobName' in event:
            return 'batch'

        return 'unknown'

    def set_baggage(self, key: str, value: str):
        """Set baggage item."""
        return baggage.set_baggage(key, value)

    def get_baggage(self, key: str) -> Optional[str]:
        """Get baggage item."""
        return baggage.get_baggage(key)

    def inject_context(self, carrier: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """Inject trace context into headers for outbound requests."""
        if carrier is None:
            carrier = {}
        propagate.inject(carrier)
        return carrier

    def extract_context(self, carrier: Dict[str, Any]) -> context.Context:
        """Extract trace context from incoming request headers."""
        return propagate.extract(carrier)

    def attach_context(self, carrier: Dict[str, Any]) -> object:
        """Extract and attach context, returning a token for cleanup."""
        ctx = self.extract_context(carrier)
        return context.attach(ctx)

    def detach_context(self, token):
        """Detach a previously attached context."""
        context.detach(token)

    @contextmanager
    def aws_message_span(self,
                         name: str,
                         message: Dict[str, Any]=None,
                         attributes: Optional[Dict[str, Any]] = None,
                         kind: SpanKind = SpanKind.CONSUMER):
        """Create span from AWS message - properly handling trace context."""

        from opentelemetry import trace, context as otel_context

        combined_attributes = attributes or {}
        combined_attributes['messaging.operation'] = 'process'

        # Extract message attributes for linking/attributes
        if message and isinstance(message, dict):
            # Add message-specific attributes
            if 'Sns' in message:
                sns_msg = message['Sns']
                if 'MessageId' in sns_msg:
                    combined_attributes['messaging.message_id'] = sns_msg['MessageId']
                if 'Subject' in sns_msg:
                    combined_attributes['messaging.sns.subject'] = sns_msg['Subject']
                if 'TopicArn' in sns_msg:
                    combined_attributes['messaging.destination'] = sns_msg['TopicArn']
                combined_attributes['messaging.system'] = 'sns'

            elif 'messageId' in message:
                # SQS message
                combined_attributes['messaging.message_id'] = message['messageId']
                if 'eventSource' in message:
                    combined_attributes['messaging.system'] = message['eventSource']


            if 'awsRegion' in message:
                combined_attributes['cloud.region'] = message['awsRegion']

            evt = extract_event_from(message)
            if evt:
                combined_attributes['event.type'] = evt


        # Use the tracer's start_span method directly to ensure it works
        # This creates a child span of whatever is currently active
        with self.tracer.start_span(
                name=name,
                attributes=combined_attributes,
                kind=kind
        ) as span:
            yield span


# Create Singleton instance
otel = RebrandlyOTEL()

# Export commonly used functions
span = otel.span
aws_message_span = otel.aws_message_span
traces = otel.trace_decorator
tracer = otel.tracer
meter = otel.meter
logger = otel.logger.logger
lambda_handler = otel.lambda_handler
aws_message_handler = otel.aws_message_handler
initialize = otel.initialize
inject_context = otel.inject_context
extract_context = otel.extract_context
attach_context = otel.attach_context
detach_context = otel.detach_context
force_flush = otel.force_flush
shutdown = otel.shutdown



# fastapi_integration.py
"""FastAPI integration for Rebrandly OTEL SDK."""
import json
from rebrandly_otel import Status, StatusCode, SpanKind
from fastapi import HTTPException, Depends
from starlette.requests import Request
from starlette.middleware.base import BaseHTTPMiddleware
from fastapi.responses import JSONResponse

import time

def setup_fastapi(otel , app):
    """
    Setup FastAPI application with OTEL instrumentation.

    Example:
        from fastapi import FastAPI
        from rebrandly_otel import otel
        from rebrandly_otel.fastapi_integration import setup_fastapi

        app = FastAPI()
        setup_fastapi(otel, app)
    """

    # Add middleware
    add_otel_middleware(otel, app)

    # Add exception handlers
    app.add_exception_handler(HTTPException, lambda request, exc: fastapi_exception_handler(otel, request, exc))
    app.add_exception_handler(Exception, lambda request, exc: fastapi_exception_handler(otel, request, exc))

    return app

def add_otel_middleware(otel, app):
    """
    Add OTEL middleware to FastAPI application.
    """

    class OTELMiddleware(BaseHTTPMiddleware):
        def __init__(self, app):
            super().__init__(app)
            self.otel = otel

        async def dispatch(self, request: Request, call_next):
            # Extract trace context from headers
            headers = dict(request.headers)
            token = self.otel.attach_context(headers)

            # Start span for request
            span_name = f"{request.method} {request.url.path}"

            # Build attributes dict, excluding None values
            attributes = {
                # Required HTTP attributes per semantic conventions
                "http.request.method": request.method,
                "http.request.headers": json.dumps(request.headers, default=str),
                "url.full": str(request.url),
                "url.scheme": request.url.scheme,
                "url.path": request.url.path,
                "network.protocol.version": "1.1",  # FastAPI/Starlette typically uses HTTP/1.1
                "server.address": request.url.hostname,
                "server.port": request.url.port or (443 if request.url.scheme == 'https' else 80),
            }

            # Add optional attributes only if they have values
            if request.url.query:
                attributes["url.query"] = request.url.query

            user_agent = request.headers.get("user-agent")
            if user_agent:
                attributes["user_agent.original"] = user_agent

            if request.client and request.client.host:
                attributes["client.address"] = request.client.host

            # Use start_as_current_span for proper context propagation
            with self.otel.tracer.tracer.start_as_current_span(
                    span_name,
                    attributes=attributes,
                    kind=SpanKind.SERVER
            ) as span:
                # Log request start
                self.otel.logger.logger.info(f"Request started: {request.method} {request.url.path}",
                                             extra={"http.method": request.method, "http.path": request.url.path})

                # Store span in request state for access in routes
                request.state.span = span
                request.state.trace_token = token

                start_time = time.time()

                try:
                    # Process request
                    response = await call_next(request)

                    # After routing, update span name and route if available
                    if hasattr(request, 'scope') and 'path' in request.scope:
                        route = request.scope.get('path', request.url.path)
                        span.update_name(f"{request.method} {route}")
                        span.set_attribute("http.route", route)

                    # Set response attributes using new semantic conventions
                    span.set_attribute("http.response.status_code", response.status_code)
                    span.set_attribute("http.status_code", response.status_code)  # Deprecated

                    # Set span status based on HTTP status code
                    if response.status_code >= 400:
                        span.set_status(Status(StatusCode.ERROR, f"HTTP {response.status_code}"))
                    else:
                        span.set_status(Status(StatusCode.OK))

                    # Log request completion
                    self.otel.logger.logger.info(f"Request completed: {response.status_code}",
                                                 extra={"http.status_code": response.status_code})
                    otel.force_flush(timeout_millis=100)
                    return response

                except Exception as e:
                    # Record exception
                    span.record_exception(e)
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.add_event("exception", {
                        "exception.type": type(e).__name__,
                        "exception.message": str(e)
                    })

                    # Log error
                    self.otel.logger.logger.error(f"Unhandled exception: {e}",
                                                  exc_info=True,
                                                  extra={"exception.type": type(e).__name__})

                    raise

                finally:
                    # Detach context
                    self.otel.detach_context(token)

    # Add middleware to app
    app.add_middleware(OTELMiddleware)

def fastapi_exception_handler(otel, request, exc):
    """
    Handle FastAPI exceptions and record them in the current span.
    """

    # Determine the status code
    if isinstance(exc, HTTPException):
        status_code = exc.status_code
        error_detail = exc.detail
    elif hasattr(exc, 'status_code'):
        status_code = exc.status_code
        error_detail = str(exc)
    elif hasattr(exc, 'code'):
        status_code = exc.code if isinstance(exc.code, int) else 500
        error_detail = str(exc)
    else:
        status_code = 500
        error_detail = str(exc)

    # Record exception in span if available and still recording
    if hasattr(request.state, 'span') and request.state.span.is_recording():
        # Update both new and old attribute names for compatibility
        request.state.span.set_attribute("http.response.status_code", status_code)
        request.state.span.set_attribute("error.type", type(exc).__name__)

        request.state.span.record_exception(exc)
        request.state.span.set_status(Status(StatusCode.ERROR, str(exc)))
        request.state.span.add_event("exception", {
            "exception.type": type(exc).__name__,
            "exception.message": str(exc)
        })

    # Log the error
    otel.logger.logger.error(f"Unhandled exception: {exc} (status: {status_code})",
                             exc_info=True,
                             extra={
                                 "exception.type": type(exc).__name__,
                                 "http.status_code": status_code
                             })

    # Return error response
    return JSONResponse(
        status_code=status_code,
        content={
            "error": error_detail,
            "type": type(exc).__name__
        }
    )

# Optional: Dependency injection helper for accessing the span in routes
def get_current_span(request: Request):
    """
    FastAPI dependency to get the current span in route handlers.

    Example:
        from fastapi import Depends
        from rebrandly_otel.fastapi_integration import get_current_span

        @app.get("/example")
        async def example(span = Depends(get_current_span)):
            if span:
                span.add_event("custom_event", {"key": "value"})
            return {"status": "ok"}
    """
    if hasattr(request.state, 'span'):
        return request.state.span
    return None


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rebrandly_otel",
    version="0.1.29",
    author="Antonio Romano",
    author_email="antonio@rebrandly.com",
    description="Python OTEL wrapper by Rebrandly",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rebrandly/rebrandly-otel-python",
    packages=["rebrandly_otel"],
    package_dir={"rebrandly_otel": "src"},
    install_requires=[
        "opentelemetry-api>=1.34.0",
        "opentelemetry-sdk>=1.34.0",
        "opentelemetry-exporter-otlp>=1.34.0",
        "psutil>=5.0.0",
        "fastapi>=0.118.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
PyMySQL instrumentation for Rebrandly OTEL SDK
Provides query tracing and slow query detection
"""

import os
import time
import functools
from opentelemetry.trace import Status, StatusCode, SpanKind

# Environment configuration
SLOW_QUERY_THRESHOLD_MS = int(os.getenv('PYMYSQL_SLOW_QUERY_THRESHOLD_MS', '1500'))
MAX_QUERY_LENGTH = 2000  # Truncate long queries


def instrument_pymysql(otel_instance, connection, options=None):
    """
    Instrument a PyMySQL connection for OpenTelemetry tracing

    Args:
        otel_instance: The RebrandlyOTEL instance
        connection: The PyMySQL connection to instrument
        options: Configuration options dict with:
            - slow_query_threshold_ms: Threshold for slow query detection (default: 1500ms)
            - capture_bindings: Include query bindings in spans (default: False for security)

    Returns:
        The instrumented connection
    """
    if options is None:
        options = {}

    slow_query_threshold_ms = options.get('slow_query_threshold_ms', SLOW_QUERY_THRESHOLD_MS)
    capture_bindings = options.get('capture_bindings', False)

    if not connection:
        print('[Rebrandly OTEL PyMySQL] No connection provided for instrumentation')
        return connection

    if not otel_instance or not hasattr(otel_instance, 'tracer'):
        print('[Rebrandly OTEL PyMySQL] No valid OTEL instance provided for instrumentation')
        return connection

    # Get tracer from RebrandlyOTEL instance
    tracer = otel_instance.tracer

    # Wrap the cursor method to return instrumented cursors
    original_cursor = connection.cursor

    def instrumented_cursor(*args, **kwargs):
        cursor = original_cursor(*args, **kwargs)
        return _instrument_cursor(cursor, tracer, slow_query_threshold_ms, capture_bindings)

    connection.cursor = instrumented_cursor

    return connection


def _instrument_cursor(cursor, tracer, slow_query_threshold_ms, capture_bindings):
    """
    Instrument a cursor's execute methods
    """
    original_execute = cursor.execute
    original_executemany = cursor.executemany

    @functools.wraps(original_execute)
    def instrumented_execute(query, args=None):
        return _trace_query(
            original_execute,
            tracer,
            slow_query_threshold_ms,
            capture_bindings,
            query,
            args,
            many=False
        )

    @functools.wraps(original_executemany)
    def instrumented_executemany(query, args):
        return _trace_query(
            original_executemany,
            tracer,
            slow_query_threshold_ms,
            capture_bindings,
            query,
            args,
            many=True
        )

    cursor.execute = instrumented_execute
    cursor.executemany = instrumented_executemany

    return cursor


def _trace_query(func, tracer, slow_query_threshold_ms, capture_bindings, query, args, many=False):
    """
    Trace a query execution with OpenTelemetry
    """
    operation = _extract_operation(query)
    truncated_query = _truncate_query(query)

    # Start span
    span_name = f"pymysql.{'executemany' if many else 'execute'}"

    with tracer.start_span(
        name=span_name,
        kind=SpanKind.CLIENT
    ) as span:
        # Set database attributes
        span.set_attribute('db.system', 'mysql')
        span.set_attribute('db.operation.name', operation)
        span.set_attribute('db.statement', truncated_query)

        # Add bindings if enabled (be cautious with sensitive data)
        if capture_bindings and args:
            if many:
                span.set_attribute('db.bindings_count', len(args))
            else:
                span.set_attribute('db.bindings', str(args))

        start_time = time.time()

        try:
            # Execute the query
            result = func(query, args)

            # Calculate duration
            duration_ms = (time.time() - start_time) * 1000
            span.set_attribute('db.duration_ms', duration_ms)

            # Check for slow query
            if duration_ms >= slow_query_threshold_ms:
                span.set_attribute('db.slow_query', True)
                span.add_event('slow_query_detected', {
                    'db.duration_ms': duration_ms,
                    'db.threshold_ms': slow_query_threshold_ms
                })

            # Set success status
            span.set_status(Status(StatusCode.OK))

            return result

        except Exception as error:
            # Calculate duration even on error
            duration_ms = (time.time() - start_time) * 1000
            span.set_attribute('db.duration_ms', duration_ms)

            # Record exception
            span.record_exception(error)
            span.set_status(Status(StatusCode.ERROR, str(error)))

            raise


def _extract_operation(sql):
    """
    Extract operation type from SQL statement

    Args:
        sql: SQL query string

    Returns:
        Operation type (SELECT, INSERT, UPDATE, etc.)
    """
    if not sql:
        return 'unknown'

    normalized = sql.strip().upper()

    if normalized.startswith('SELECT'):
        return 'SELECT'
    if normalized.startswith('INSERT'):
        return 'INSERT'
    if normalized.startswith('UPDATE'):
        return 'UPDATE'
    if normalized.startswith('DELETE'):
        return 'DELETE'
    if normalized.startswith('CREATE'):
        return 'CREATE'
    if normalized.startswith('DROP'):
        return 'DROP'
    if normalized.startswith('ALTER'):
        return 'ALTER'
    if normalized.startswith('TRUNCATE'):
        return 'TRUNCATE'

    return 'unknown'


def _truncate_query(sql):
    """
    Truncate long queries for span attributes

    Args:
        sql: SQL query string

    Returns:
        Truncated query
    """
    if not sql:
        return ''
    if len(sql) <= MAX_QUERY_LENGTH:
        return sql
    return sql[:MAX_QUERY_LENGTH] + '... [truncated]'
